# Copyright (c) 2025 Oleg Anufriev
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from contextlib import contextmanager
import os
import re
import shlex
import signal
from subprocess import Popen, PIPE, STDOUT
import sys
from typing import Generator, Union, Dict, List, IO


DO_APPEND_RE = re.compile(r"^\+")


def open_stream_w(path: Union[str, os.PathLike] | None) -> IO | None:
    """
    Open/create file for write/append.

    :param path: file name
    """

    if path is None:
        return None
    else:
        mode: str = "a" if DO_APPEND_RE.match(str(path)) is not None else "w"
        return open(re.sub(r"^\+", "", str(path)).strip(), mode)


def open_stream_r(path: Union[str, os.PathLike] | None) -> IO | None:
    """
    Open file for reading.

    :param path: file name
    """

    if path is None:
        return None
    else:
        return open(re.sub(r"^\+", "", str(path)).strip(), "r")


@contextmanager
def prepare_std(
    inptxt: str = None,
    inpfl: Union[str, os.PathLike] = None,
    outfl: Union[str, os.PathLike] = None,
    errfl: Union[str, os.PathLike] = None,
    err2out: bool = False,
) -> Generator:
    """
    Prepare pipeline std streams.

    :param inptxt:  text to be written to pipeline stdin
    :param inpfl:   file name to connect to stdin; exclusive mutual with inptxt
    :param outfl:   file name to redirect stdout to
    :param errfl:   file name to redirect stderr to; exclusive mutual with err2out
    :param err2out: redirect stderr to stdout; exclusive mutual with errfl
    :yields:        tuple of std streams
    """

    stdout: IO = open_stream_w(outfl)

    if inptxt is not None and inpfl is not None:
        raise Exception("Arguments conflict: ['inptxt', 'inpfl']")

    if inptxt is not None:
        stdin: int = PIPE
    else:
        stdin = open_stream_r(inpfl)

    if err2out:
        stderr: int = STDOUT
        if errfl is not None:
            raise Exception("Arguments conflict: ['errfl', 'err2out']")
    else:
        stderr: IO = open_stream_w(errfl)

    try:
        yield (stdin, stdout, stderr)
    finally:
        map(lambda stream: stream.close(), filter(lambda stream: stream is not None, [stdin, stdout, stderr]))


@contextmanager
def wrap_tty() -> Generator:
    """
    Deal with process group to control terminal association

    * memorize current process group associated to control terminal
    * inhibit SIGTTOU handling
    * roll back things afterwards

    :yields: control tty file descriptor
    """
    tty_path: str = os.ttyname(sys.stdin.fileno())
    tty_fd: IO = os.open(tty_path, os.O_RDONLY)

    orig_tcpgrp: int = os.tcgetpgrp(tty_fd)
    orig_sigttou_handler = signal.getsignal(signal.SIGTTOU)
    signal.signal(signal.SIGTTOU, signal.SIG_IGN)
    try:
        yield tty_fd
    finally:
        os.tcsetpgrp(tty_fd, orig_tcpgrp)
        signal.signal(signal.SIGTTOU, orig_sigttou_handler)
        os.close(tty_fd)


def construct_pipeline(*cmds: str, cwd=None, env=None, stdin=None, stdout=None, stderr=None) -> List[Popen]:
    """
    Spawn processes, connect them by pipes, connect std streams.

    :param *cmds:  pipeline commands
    :param cwd:    pipeline working directory
    :param env:    extra envrionment variables to set for pipeline
    :param stdin:  input stream
    :param stdout: output stream
    :param stderr: error stream
    :returns: list of pipeline Popen objects
    """

    prcs: List[Popen] = []
    pipeline_size: int = len(cmds)
    pgid: int = 0

    prcs_env: Dict = os.environ.copy()
    if env is not dict:
        prcs_env.update(env)

    for i, cmd in enumerate(cmds):
        cmd_as_list: List[str] = shlex.split(cmd)

        prc_stdin: IO | int | None = None
        prc_stdout: IO | int | None = None
        prc_stderr: IO | int | None = stderr

        if i == 0:
            prc_stdin = stdin
        if i > 0:
            prc_stdin = prcs[i - 1].stdout

        if i < pipeline_size - 1:
            prc_stdout = PIPE
        if i == pipeline_size - 1:
            prc_stdout = stdout

        prc = Popen(
            cmd_as_list,
            stdin=prc_stdin,
            stdout=prc_stdout,
            stderr=prc_stderr,
            env=env,
            bufsize=1,
            encoding="utf-8",
            text=True,
            process_group=pgid,
            cwd=cwd,
        )
        if pgid == 0:
            pgid = prc.pid

        prcs.append(prc)

    return prcs



# Copyright (c) 2025 Oleg Anufriev
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import os
from signal import SIGKILL, SIGTERM
from subprocess import TimeoutExpired, Popen
import time
from threading import Thread
from typing import Self, List, Set


class Task:
    """
    Representer of pipeline running in background.

    :param prcs: list of pipeline Popen objects
    """

    def __init__(self, prcs: List[Popen]) -> Self:
        self._prcs: List[Popen] = prcs
        self.pid: int = prcs[0].pid

    @property
    def returncode(self) -> int | None:
        """
        Pipeline retirncode.

        :returns: pipeline returncode if one is complete, else None
        """
        return self._prcs[-1].poll()

    @property
    def is_alive(self) -> bool:
        """
        Is pipeline still running.
        """

        rc: int | None = self._prcs[-1].poll()
        if rc is None:
            return True

        return False

    def __repr__(self) -> str:
        cmds: List[str] = []
        for prc in self._prcs:
            cmds.append(" ".join(prc.args))
        return f"""Task(cmd="{" | ".join(cmds)}", alive={self.is_alive}, pid={self.pid})"""

    def kill(self, signal: int = SIGTERM) -> None:
        """
        Send signal to pipeline.
        """
        try:
            os.killpg(self.pid, signal or SIGKILL)
        except ProcessLookupError:
            pass

    def wait(self, timeout: int = None) -> int:
        """
        Wait for pipeline to complete.

        :timeout:                         pipeline completion wait timeout
        :returns:                         pipeline return code
        :raise subprocess.TimeoutExpired: on waiting timeout expired
        """

        for prc in self._prcs:
            try:
                return prc.wait(timeout)
            except TimeoutExpired:
                self.kill(signal=SIGKILL)
                raise


class TaskPoller:
    def __init__(self) -> Self:
        self.tasks: Set = set()
        self._poll_th: Thread | None = None

    def _poll_func(self) -> None:
        while True:
            finished_tasks: Set = set()
            time.sleep(0.1)
            if not self.tasks:
                break
            for task in self.tasks:
                rc: int | None = task.returncode
                if rc is not None:
                    finished_tasks.add(task)

            self.tasks.difference_update(finished_tasks)

    def add(self, task: Task) -> Task:
        """
        Add task for polling.
        """

        self.tasks.add(task)
        if self._poll_th is None or not self._poll_th.is_alive():
            self._poll_th = Thread(target=self._poll_func)
            self._poll_th.start()

        return task

    def __del__(self) -> None:
        for task in self.tasks:
            task.kill(signal=SIGKILL)



# Copyright (c) 2025 Oleg Anufriev
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import copy
import os
import signal
from subprocess import PIPE, CalledProcessError, Popen
from typing import Union, Self, LiteralString, Dict, List

from piperum.background import TaskPoller, Task
from piperum import utils


class Piperum:
    """
    Handy subprocess pipelines runner with pipelines support.

    :param task_poller: background.TaskPoller; polls background tasks completion
    """

    def __init__(self, task_poller: TaskPoller) -> Self:
        self._env: Dict[LiteralString, LiteralString] = {}
        self._cwd: Union[str, os.PathLike] | None = None
        self._task_poller: TaskPoller = task_poller

    def __call__(
        self,
        cwd: os.PathLike | None = None,
        **env: LiteralString,
    ) -> Self:
        """
        Change.

        :param cwd:   pipelines working directory
        :param **env: environment variables
        """

        if not all(map(lambda x: type(x) is str, env.values())):
            raise TypeError("env variable value should be of type str.")

        new_piperum = copy.copy(self)

        new_piperum._cwd = cwd if cwd is not None else self._cwd

        new_piperum._env = self._env
        new_piperum._env.update(env)

        return new_piperum

    def __enter__(self) -> Self:
        return self

    def __exit__(self, *_) -> None:
        pass

    def _kill(self, prcs: List[Popen]) -> None:
        """
        Send signal to process group.

        :param prcs: list of pipeline processes
        """

        try:
            os.killpg(prcs[0].pid, signal.SIGKILL)
        except ProcessLookupError:
            pass
        map(lambda prc: prc.wait(), prcs)

    def run(
        self,
        *cmds: str,
        inptxt: str | None = None,
        inpfl: Union[str, os.PathLike] | None = None,
        outfl: Union[str, os.PathLike] | None = None,
        errfl: Union[str, os.PathLike] | None = None,
        err2out: bool = False,
        timeout: int | None = None,
    ) -> None:
        """
        Run pipeline in foreground.

        :param *cmds:                           pipeline commands
        :param inptxt:                          text to be written to pipeline stdin; exclusive mutual with inpfl
        :param inpfl:                           file name to connect to stdin; exclusive mutual with inptxt
        :param outfl:                           file name to redirect stdout to
        :param errfl:                           file name to redirect stderr to; exclusive mutual with err2out
        :param err2out:                         redirect stderr to stdout; exclusive mutual with errfl
        :param timeout:                         pipeline completion wait timeout
        :raises subprocess.CalledProcessError:  any process of pipeline return code != 0
        """

        prc_inp: Union[str, os.PathLike] | None = inptxt or inpfl
        retcode: int = 0

        with (
            utils.wrap_tty() as tty_fd,
            utils.prepare_std(inptxt=inptxt, inpfl=inpfl, outfl=outfl, errfl=errfl, err2out=err2out) as std,
        ):
            prcs: List[Popen] = utils.construct_pipeline(
                *cmds, cwd=self._cwd, env=self._env, stdin=std[0], stdout=std[1], stderr=std[2]
            )

            os.tcsetpgrp(tty_fd, prcs[0].pid)

            try:
                for prc in prcs:
                    if prc_inp is None:
                        prc.wait(timeout)
                    else:
                        prc.communicate(input=prc_inp, timeout=timeout)
                        prc_inp = None

                    retcode = prc.returncode

                    if retcode:
                        raise CalledProcessError(retcode, prc.args)
            finally:
                self._kill(prcs)

    def cap(
        self,
        *cmds: str,
        inptxt: str | None = None,
        inpfl: Union[str, os.PathLike] | None = None,
        errfl: Union[str, os.PathLike] | None = None,
        err2out=False,
        timeout: int | None = None,
    ) -> str | None:
        """
        Run pipeline in foreground and capture stdout.

        :param *cmds:                           pipeline commands
        :param inptxt:                          text to be written to pipeline stdin; exclusive mutual with inpfl
        :param inpfl:                           file name to connect to stdin; exclusive mutual with inptxt
        :param errfl:                           file name to redirect stderr to; exclusive mutual with err2out
        :param err2out:                         redirect stderr to stdout; exclusive mutual with errfl
        :param timeout:                         pipeline completion wait timeout
        :returns:                               captured text
        :raises subprocess.CalledProcessError:  any process of pipeline return code != 0
        """

        prc_inp: str | None = inptxt or inpfl
        res_out: str = None
        retcode: int = 0

        with (
            utils.wrap_tty() as tty_fd,
            utils.prepare_std(inptxt=inptxt, inpfl=inpfl, outfl=None, errfl=errfl, err2out=err2out) as std,
        ):
            prcs: List[Popen] = utils.construct_pipeline(
                *cmds, cwd=self._cwd, env=self._env, stdin=std[0], stdout=PIPE, stderr=std[2]
            )

            os.tcsetpgrp(tty_fd, prcs[0].pid)

            try:
                prcs_len: int = len(prcs)
                res_out: Union[str, bytes] = ""
                for i, prc in enumerate(prcs):
                    if prcs_len == 1:
                        res_out, _ = prc.communicate(input=prc_inp, timeout=timeout)
                    else:
                        if prc_inp is not None:
                            prc.communicate(input=prc_inp, timeout=timeout)
                            prc_inp = None
                        elif i == prcs_len - 1:
                            res_out, _ = prc.communicate(timeout=timeout)
                        else:
                            prc.wait(timeout)

                    retcode = prc.returncode

                    if retcode:
                        raise CalledProcessError(retcode, prc.args)
            finally:
                self._kill(prcs)

        return res_out

    def run_bg(
        self,
        *cmds: str,
        inpfl: Union[str, os.PathLike] | None = None,
        outfl: Union[str, os.PathLike] | None = None,
        errfl: Union[str, os.PathLike] | None = None,
        err2out: bool = False,
    ) -> Task | None:
        """
        Run pipeline in background.

        :param *cmds:   pipeline commands
        :param inpfl:   file name to connect to stdin
        :param outfl:   file name to redirect stdout to
        :param errfl:   file name to redirect stderr to; exclusive mutual with err2out
        :param err2out: redirect stderr to stdout; mutual exclusive with errfl
        :returns:       background.Task
        """

        with utils.prepare_std(inpfl=inpfl, outfl=outfl, errfl=errfl, err2out=err2out) as std:
            prcs: List[Popen] = utils.construct_pipeline(
                *cmds, cwd=self._cwd, env=self._env, stdin=std[0], stdout=std[1], stderr=std[2]
            )

        return self._task_poller.add(Task(prcs))
